import wrds
from datetime import datetime
import os

def getCRSPData(params):
    username = params.username
    crsp_csv_file_path = params.crspFolder + os.sep
    start = params.sample_start
    end = params.sample_end
    "Establish connection to WRDS"
    db = wrds.Connection(wrds_username=username)

    "Timekeeping"
    print(f"\nNow working on collecting data from CRSP. Run started at {datetime.now()}.\n")

    """Want to download and save the following tables from CRSP:
     MSFHDR, MSF, MSEDELIST, MSEEXCHDATES, CCMXPF_LNKHIST, STOCKNAMES"""

    crsp_msfhdr = db.raw_sql("select * from CRSP.MSFHDR")
    crsp_msfhdr.to_csv(crsp_csv_file_path + 'crsp_msfhdr.csv')

    crsp_msf = db.raw_sql(f"select * from CRSP.MSF where date>='01/01/{start}' and date<='12/31/{end}'")

    crsp_msf.to_csv(crsp_csv_file_path + 'crsp_msf.csv')

    crsp_msedelist = db.raw_sql("select * from CRSP.MSEDELIST")
    crsp_msedelist.to_csv(crsp_csv_file_path + 'crsp_msedelist.csv')

    crsp_mseexchdates = db.raw_sql("select * from CRSP.MSEEXCHDATES")
    crsp_mseexchdates.to_csv(crsp_csv_file_path + 'crsp_mseexchdates.csv')

    crsp_ccmxpf_lnkhist = db.raw_sql("select * from CRSP.CCMXPF_LNKHIST")
    crsp_ccmxpf_lnkhist.to_csv(crsp_csv_file_path + 'crsp_ccmxpf_lnkhist.csv')

    crsp_stocknames = db.raw_sql("select * from CRSP.STOCKNAMES")
    crsp_stocknames.to_csv(crsp_csv_file_path + 'crsp_stocknames.csv')

    "Timekeeping"
    print(f"\nFinished collecting data from CRSP. Run ended at {datetime.now()}.\n")

    db.close()

# import wrds


#
# "Set path"
# crspFolder = r'/home/jlaws13/PycharmProjects/AssayingAnomalies_root/Data/CRSP/'
#
# "Establish connection to WRDS"
# db = wrds.Connection(wrds_username='jlaws13')
#
# """Want to download and save the following tables from CRSP:
#  MSFHDR, MSF, MSEDELIST, MSEEXCHDATES, CCMXPF_LNKHIST, STOCKNAMES"""
#
#
# crsp_msfhdr = db.raw_sql("select * from CRSP.MSFHDR")
# crsp_msfhdr.to_csv(crspFolder + 'crsp_msfhdr.csv')
#
# crsp_msf = db.raw_sql("select * from CRSP.MSF")
# crsp_msf.to_csv(r'C:\Users\josh\OneDrive - University at Buffalo\Desktop\Spring_2023\Empirical_Asset_Pricing\AssayingAnomalies\Data\crsp_msf.csv')
#
# crsp_msedelist = db.raw_sql("select * from CRSP.MSEDELIST")
# crsp_msedelist.to_csv(r'C:\Users\josh\OneDrive - University at Buffalo\Desktop\Spring_2023\Empirical_Asset_Pricing\AssayingAnomalies\Data\crsp_msedelist.csv')
#
# crsp_mseexchdates = db.raw_sql("select * from CRSP.MSEEXCHDATES")
# crsp_mseexchdates.to_csv(r'C:\Users\josh\OneDrive - University at Buffalo\Desktop\Spring_2023\Empirical_Asset_Pricing\AssayingAnomalies\Data\crsp_mseexchdates.csv')
#
# crsp_ccmxpf_lnkhist = db.raw_sql("select * from CRSP.CCMXPF_LNKHIST")
# crsp_ccmxpf_lnkhist.to_csv(r'C:\Users\josh\OneDrive - University at Buffalo\Desktop\Spring_2023\Empirical_Asset_Pricing\AssayingAnomalies\Data\crsp_ccmxpf_lnkhist.csv')
#
# crsp_stocknames = db.raw_sql("select * from CRSP.STOCKNAMES")
# crsp_stocknames.to_csv(r'C:\Users\josh\OneDrive - University at Buffalo\Desktop\Spring_2023\Empirical_Asset_Pricing\AssayingAnomalies\Data\crsp_stocknames.csv')
#
# db.close()




import requests
import pandas as pd
from datetime import datetime


def getFredData(series_id, observation_start=None, observation_end=None, units=None,
                frequency=None, aggregation_method=None, ondate=None, realtime_end=None):
    """
    Fetches and returns economic data from the FRED (Federal Reserve Economic Data) API.

    This function queries the FRED API to obtain time series data for a specified economic data series.
    It allows for various optional parameters to refine the data query, such as observation start and end dates,
    units, frequency, and aggregation methods. The function returns a pandas DataFrame with the requested data.

    Parameters:
    series_id (str): The FRED series ID for which data is requested.
    observation_start (str, optional): The start date for the data observations (format: 'YYYY-MM-DD').
    observation_end (str, optional): The end date for the data observations (format: 'YYYY-MM-DD').
    units (str, optional): The data value transformations (e.g., 'lin', 'chg', 'ch1', 'pch', 'pc1', 'pca').
    frequency (str, optional): Data frequency ('d', 'w', 'bw', 'm', 'q', 'sa', 'a', 'wef', 'weth', 'wew', 'wetu', 'wem', 'wesu', 'semi', 'wew', 'weq', 'ag').
    aggregation_method (str, optional): Aggregation method used for frequency aggregation ('avg', 'sum', 'eop').
    ondate (str, optional): The date for which the real-time period begins (format: 'YYYY-MM-DD').
    realtime_end (str, optional): The date for which the real-time period ends (format: 'YYYY-MM-DD'). If not provided, 'ondate' will be used.

    Returns:
    pandas.DataFrame: A DataFrame containing the requested time series data. The DataFrame is indexed by date, and
                      contains columns for the date, observation value, and other metadata related to each observation.

    Raises:
    HTTPError: An error is raised if the API request fails.

    Example:
    series_id = 'CPIAUCNS'
    df = getFredData(series_id, observation_start='2000-01-01', observation_end='2020-01-01')
    """

    api_key = '54fa5832fc9d049bff2f59e7f1b4064e'
    api_key_mish = '625723b3573b7c847f626c6df4254c03'
    base_url = 'https://api.stlouisfed.org/fred/series/observations'

    # Construct the query parameters
    query_params = {
        'series_id': series_id,
        'api_key': api_key,
        'file_type': 'json'  # Get data in JSON format
    }
    # Add optional parameters if provided
    if observation_start:
        query_params['observation_start'] = observation_start
    if observation_end:
        query_params['observation_end'] = observation_end
    if units:
        query_params['units'] = units
    if frequency:
        query_params['frequency'] = frequency
    if aggregation_method:
        query_params['aggregation_method'] = aggregation_method
    if ondate:
        query_params['realtime_start'] = ondate
        query_params['realtime_end'] = ondate if not realtime_end else realtime_end

    # Make the API request
    response = requests.get(base_url, params=query_params)
    response.raise_for_status()  # Raise an exception for HTTP errors

    # Parse the response data
    data = response.json()['observations']
    df = pd.DataFrame(data)

    # Convert date strings to datetime objects and set as index
    df['date'] = pd.to_datetime(df['date'])
    df.set_index('date', inplace=True)

    # Convert value strings to numeric, handling missing values
    df['value'] = pd.to_numeric(df['value'], errors='coerce')

    return df

# testing = getFredData('CPIAUCNS', observation_start='2003-01-01', observation_end='2023-01-01')




# config.py
import os
import json
import shutil
import pkg_resources
import os


class Config:
    def __init__(self):
        self.username = None
        self.password = None
        self.sample_start = None
        self.sample_end = None
        self.domComEqFlag = None
        self.data_storage_path = None
        self.data_folder = None
        self.crspFolder = None
        self.compFolder = None
        self.daily_crsp_folder = None
        self.ff_data_folder = None
        self.num_cpus = os.cpu_count()
        self.gibbs_data_folder = None
        self.hf_effective_spreads_folder = None

    def prompt_user(self):
        self.username = str(input("Enter your WRDS username: "))
        self.password = str(input("Enter your WRDS password: "))
        self.sample_start = int(input("Enter sample start year: "))
        self.sample_end = int(input("Enter sample end year: "))

        # For Boolean input, check the string entered
        domComEqInput = input("Enter True if you would like only domestic common equity otherwise False: ")
        self.domComEqFlag = domComEqInput.strip().lower() == 'true'
        remoteInput = input("Enter True if you are using BlueHive or CIRC, otherwise False: ")
        self.remote_or_not = remoteInput.strip().lower() == 'true'

        self.data_storage_path = input("Enter the path to where you would like the downloaded data and variables"
                                       " stored. If you are using CIRC and are unsure, enter /scratch/user_id: ")

        # if self.remote_or_not:
        #     self.num_cpus = int(input("Enter the number of CPUs. If unsure, enter 1: "))

    def make_folders(self):
        paths = []
        self.data_folder = os.path.join(self.data_storage_path, 'AA_Data')
        paths.append(self.data_folder)
        self.crspFolder = os.path.join(self.data_folder, 'CRSP')
        paths.append(self.crspFolder)
        self.compFolder = os.path.join(self.data_folder, 'COMPUSTAT')
        paths.append(self.compFolder)
        self.daily_crsp_folder = os.path.join(self.crspFolder, 'Daily')
        paths.append(self.daily_crsp_folder)
        self.ff_data_folder = os.path.join(self.data_folder, 'FFData')
        paths.append(self.ff_data_folder)
        self.gibbs_data_folder = os.path.join(self.data_folder, 'Gibbs')
        paths.append(self.gibbs_data_folder)
        self.hf_effective_spreads_folder = os.path.join(self.data_folder, 'High-frequency effective spreads')
        paths.append(self.hf_effective_spreads_folder)
        for path in paths:
            os.makedirs(path, exist_ok=True)

    def save_params(self, filepath="config.json"):
        with open(filepath, 'w') as f:
            # Convert configuration parameters to a dictionary and save as JSON
            config_dict = self.__dict__
            json.dump(config_dict, f)

    def move_trading_costs_files(self):
        # Define the source folders
        gibbs_source = pkg_resources.resource_filename('AssayingAnomalies', 'Gibbs')
        hf_spreads_source = pkg_resources.resource_filename('AssayingAnomalies', 'High-frequency effective spreads')

        # Define a dictionary mapping source folders to target folders
        source_to_target = {
            gibbs_source: self.gibbs_data_folder,
            hf_spreads_source: self.gibbs_data_folder
        }

        # Move files from each source folder to the respective target folder
        for source, target in source_to_target.items():
            if os.path.exists(source):
                for filename in os.listdir(source):
                    shutil.move(os.path.join(source, filename), os.path.join(target, filename))


    @staticmethod
    def load_params(filepath="config.json"):
        if os.path.exists(filepath):
            with open(filepath, 'r') as f:
                config_dict = json.load(f)
                config = Config()
                config.__dict__.update(config_dict)
                return config
        else:
            return None

    def set_up(self):
        self.prompt_user()
        self.make_folders()
        self.move_trading_costs_files()
        self.save_params()


# test = Config()
# test.prompt_user()




import pandas as pd
import numpy as np
import os
from datetime import datetime

from AssayingAnomalies.wrds_utilities.utilities import ssh_login, start_qrsh, start_python_wrds, exit_sessions,\
    execute_commands, transfer_files, delete_files


def get_hf_spreads_data(params, **kwargs):
    print(f"Now retrieving HF effective spreads data. Run started at {datetime.now()}")

    # Default keyword arguments
    p = {
        "delete_files_ok": True
    }

    # Update keyword arguments with user-specified values (if provided).
    for key, value in kwargs.items():
        if key in p:
            p[key] = value

    # Define the columns to pull from WRDS columns_to_pull = ['SYM_ROOT', 'SYM_SUFFIX', 'symbol', 'DATE',
    # 'quotedspread_percent_tw', 'effectivespread_percent_dw']  # The Chen-Velikov SAS code pulls these measures (and
    # others) but the Assaying anomalies package only uses the effective spreads, not quoted spreads.
    columns_to_pull = ['SYM_ROOT', 'SYM_SUFFIX', 'symbol', 'DATE', 'effectivespread_percent_dw']

    # Start a new child process, and connect to the SSH server
    child = ssh_login(params)

    # Start a qrsh session
    start_qrsh(params, child)

    # Start Python session
    start_python_wrds(child)

    # Create the save_file names and execute commands.
    date_range = np.unique(np.linspace(max(params.sample_start, 2003), params.sample_end, max(params.sample_end, 2003) - params.sample_start + 1).astype(int))
    commands = []
    file_locations = []
    for year in date_range:
        save_name = 'spreads_'
        sql_string = "SELECT " + ", ".join(columns_to_pull) + f" FROM taqmsec.wrds_iid_{year}"
        command1 = f"testing = db.raw_sql(\"{sql_string}\")"
        file_name = f"{save_name}{year}.csv"
        command2 = f"testing.to_csv(\"/scratch/rochester/{file_name}\")"
        commands.extend([command1, command2])
        file_location = f'/scratch/rochester/{file_name}'
        file_locations.append(file_location)

    # Execute python commands on wrds cloud
    execute_commands(child, commands)

    # Exit all sessions
    exit_sessions(child)

    # Transfer the files and return a list of files that did not transfer
    print("Using SFTP to transfer files.")
    save_folder = params.hf_effective_spreads_folder
    problem_files = transfer_files(params, file_locations, save_folder)

    # Delete the files
    if p['delete_files_ok']:
        delete_files(params, file_locations, problem_files)

    # Print statement indicating if all files were completely transferred or not.
    if problem_files:
        print("The following files did not transfer: ")
        for file in problem_files:
            print(file)
    else:
        print("All files were successfully transferred and deleted from WRDS storage.")

    print(f"Finished retrieving HF effective spreads data. Run ended at {datetime.now()}")
    return


def make_hf_effective_spreads(params):
    print(f"Now making HF effective spreads. Run started at {datetime.now()}")

    hf_path = params.hf_effective_spreads_folder + os.sep
    crsp_path = params.crspFolder + os.sep

    # Initialize dataframes
    hf_spreads_ave = pd.DataFrame()
    hf_spreads_last = pd.DataFrame()

    # Load some variables
    msenames = pd.read_csv(crsp_path + 'crsp_stocknames.csv', index_col=0)
    ret = pd.read_csv(params.crspFolder + os.sep + 'ret.csv', index_col=0).astype(float)

    date_range = [i for i in range(max(params.sample_start, 2003), params.sample_end+1)]
    for year in date_range:
        data = pd.read_csv(hf_path + f'spreads_{year}.csv', index_col=0)
        # q_spread = pd.pivot(data, columns='symbol', index='date', values='quotedspread_percent_tw')
        e_spread = pd.pivot(data, columns='symbol', index='date', values='effectivespread_percent_dw')
        e_spread.index = pd.to_datetime(e_spread.index, format='%Y-%m-%d')

        # Setting index to datetime
        e_spread_ave = e_spread.resample('M').mean()
        e_spread_last = e_spread.resample('M').last()

        # Create a mapping from tickers to permno.
        ticker_to_permno = msenames.set_index('ticker')['permno'].to_dict()

        # Identify columns in e_spread that have a corresponding mapping in msenames
        mapped_columns = [col for col in e_spread_ave.columns if col in ticker_to_permno]

        # Restrict e_spread to these columns
        e_spread_ave = e_spread_ave[mapped_columns]
        e_spread_last = e_spread_last[mapped_columns]

        # Map the column names to permno if needed
        e_spread_ave.columns = [ticker_to_permno[col] for col in e_spread_ave.columns]
        e_spread_last.columns = [ticker_to_permno[col] for col in e_spread_last.columns]

        # Convert the columns of e_spread to float
        e_spread_ave.columns = e_spread_ave.columns.astype(str)
        e_spread_last.columns = e_spread_last.columns.astype(str)
        # TODO: change the below portion to .reindex, it is much faster
        # Add NaN columns to e_spread for any columns in ret that are not in e_spread
        for col in ret.columns:
            if col not in e_spread_ave.columns:
                e_spread_ave[col] = np.nan
            if col not in e_spread_last:
                e_spread_last[col] = np.nan

        # Remove duplicates by keepiing the first occurence. Copying Zhen and Velikov here
        e_spread_ave = e_spread_ave.iloc[:, ~e_spread_ave.columns.duplicated(keep='first')]
        e_spread_last = e_spread_last.iloc[:, ~e_spread_last.columns.duplicated(keep='first')]
        e_spread_ave = e_spread_ave[np.intersect1d(ret.columns, e_spread_ave.columns)]
        e_spread_last = e_spread_last[np.intersect1d(ret.columns, e_spread_last.columns)]

        # Append
        hf_spreads_ave = hf_spreads_ave.append(e_spread_ave)
        hf_spreads_last = hf_spreads_last.append(e_spread_last)

    # Save dataframes
    hf_spreads_ave.to_csv(hf_path + 'hf_spreads_ave_post_2003.csv')
    hf_spreads_last.to_csv(hf_path + 'hf_spreads_last_post_2003.csv')

    print(f"Finished making HF effective spreads. Run ended at {datetime.now()}")

    return


def extend_hf_effective_spreads(params):
    print(f"Extending HF effective spreads to pre-2003. Run started at {datetime.now()}")
    hf_path = params.hf_effective_spreads_folder + os.sep
    if params.sample_start <= 2003:
        # Extend using the csv created after running the HF Effective Spreads SAS code.
        pre_2003 = pd.read_csv(hf_path + 'hf_monthly_pre_2003.csv', low_memory=False)
        pre_2003 = pd.pivot(pre_2003, columns='permno', values='espread_pct_mean', index='yearm')
        pre_2003.columns = pre_2003.columns.astype(str)
        pre_2003.index = pd.to_datetime(pre_2003.index, format='%Y%m')
        pre_2003.index = pre_2003.index.to_period('M')

        # Load and filter data from make_hf_effective_spreads
        post_2003 = pd.read_csv(hf_path + 'hf_spreads_ave_post_2003.csv', index_col=0)
        post_2003.index = pd.to_datetime(post_2003.index)
        post_2003.index = post_2003.index.to_period('M')
        post_2003 = post_2003[post_2003.index >= pd.Period('2004-01')]

        # Make sure the pre_2003 data has the same columns as the post_2003 df
        pre_2003 = pre_2003.reindex(columns=post_2003.columns, fill_value=np.nan)
        pre_2003 = pre_2003 / 100

        # Select rows before 2003
        start_date = pd.Period(str(params.sample_start) + '-01')
        end_date = pd.Period(str(min(2003, params.sample_end)) + '-12')
        pre_2003 = pre_2003.loc[(start_date <= pre_2003.index) & (pre_2003.index <= end_date)]

        # Combine the dataframes.
        combined_df = pd.concat([pre_2003, post_2003])

        # Save the dataframe
        combined_df.to_csv(hf_path + 'hf_monthly.csv')

        print(f"Finished extending hf-effective spreads to pre-2003. Run ended at {datetime.now()}")

        return combined_df

    else:
        # Load and filter data from make_hf_effective_spreads
        post_2003 = pd.read_csv(hf_path + 'hf_spreads_ave_post_2003.csv', index_col=0)
        post_2003.index = pd.to_datetime(post_2003.index)
        post_2003.index = post_2003.index.to_period('M')
        post_2003 = post_2003[post_2003.index >= pd.Period('2004-01')]

        print(f"Finished extending hf-effective spreads to pre-2003. Run ended at {datetime.now()}")

        return post_2003


# get_hf_spreads_data(params)
# make_hf_effective_spreads(params)
# extend_hf_effective_spreads(params)



from setuptools import setup, find_packages

# Read requirements.txt and store its contents in the 'requirements' variable
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

# Read the content of REAMDE file.
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='AssayingAnomalies',
    version='1.5.2',
    author='Joshua Lawson',
    author_email='jlaws13@simon.rochester.edu',
    description='This library is a Python implementation of the MATLAB Toolkit that accompanies Novy-Marx and Velikov '
                '(2023) and is to be used for empirical academic asset pricing research, particularly focused on '
                'studying anomalies in the cross-section of stock returns.',
    packages=find_packages(),
    package_data={
        'AssayingAnomalies': ['Gibbs/*', 'High-frequency effective spreads/*'],
    },
    install_requires=requirements,
    # long_description='Package installed successfully. Please run \'setup_library\' to configure your settings and '
    #                  'begin using the toolkit'
    long_description=long_description,
    long_description_content_type="text/markdown"
)




import wrds
import os
from datetime import datetime
import numpy as np


# :TODO Test the .parquet file download and transfer.
def get_crsp_daily_data(params):

    username = params.username
    crsp_csv_file_path = params.daily_crsp_folder + os.sep

    "Timekeeping"
    print(f"\nNow working on collecting data from CRSP. Run started at {datetime.now()}.\n")

    # Define the columns to pull from WRDS
    columns_to_pull = ['permno', 'date', 'cfacpr', 'cfacshr', 'bidlo', 'askhi', 'prc', 'vol', 'ret', 'bid',
                       'ask', 'shrout', 'openprc', 'numtrd']

    # Set the beginning and ending year
    start = params.sample_start
    end = params.sample_end

    # Manually select years before 2000 based on your intervals
    years_before_2000 = [start, 2000 - 55, 2000 - 25, 2000 - 10]  # 1925, 1970, 1985, 1990

    # Ensure years_before_2000 are within the range and unique
    years_before_2000 = [year for year in years_before_2000 if start <= year < 2000]
    years_before_2000 = sorted(set(years_before_2000))

    # Years from 2000 to end, split into three periods
    if end >= 2000:
        years_2000_to_end = np.linspace(2000, end, 3, endpoint=True).astype(int)
    else:
        years_2000_to_end = []

    # Combine and sort the lists
    years = sorted(set(years_before_2000 + years_2000_to_end.tolist()))

    "Establish connection to WRDS"
    db = wrds.Connection(wrds_username=username)
    # Create commands for custom sql query and saving files

    for start, end in zip(years, years[1:]):
        # dsf_file_name = f"crsp_dsf_{end}.csv"
        dsf_file_name = f"crsp_dsf_{end}.parquet"
        dsf_sql_string = "SELECT " + ", ".join(columns_to_pull) + f" FROM CRSP.DSF " \
                f"WHERE date>='01-01-{start}' " \
                f"and date<'01-01-{end+1}' "
        temp = db.raw_sql(dsf_sql_string)
        temp.to_csv(crsp_csv_file_path + dsf_file_name)

    # Create the delist file names, location, and commands
    delist_file_name = "crsp_dsedelist.csv"
    delist_sql_string = "SELECT * FROM CRSP.DSEDELIST"
    temp = db.raw_sql(delist_sql_string)
    temp.to_csv(crsp_csv_file_path + delist_file_name)

    "Timekeeping"
    print(f"\nFinished collecting daily data from CRSP. Run ended at {datetime.now()}.\n")

    db.close()

    return




import numpy as np
import pandas as pd
import requests
from io import BytesIO
import os
import zipfile

# crspFolder = r'/home/jlaws13/PycharmProjects/AssayingAnomalies_root/Data/CRSP/'

def getFFFactors(params):
    """

    :return:
    """
    url = 'http://mba.tuck.dartmouth.edu/pages/faculty/ken.french/ftp/F-F_Research_data_Factors_CSV.zip'

    # Set path to CRSP Folder
    crspFolder = params.crspFolder + os.sep

    # Load dates
    dates = pd.read_csv(crspFolder + 'dates.csv', index_col=0)

    # tricks the host site into thinking the HTTP call is from user instead of machine
    headers = {"User-Agent": "Mozilla/5.0"}
    response = requests.get(url, headers=headers)
    # saving the zip file as an object
    zf = zipfile.ZipFile(BytesIO(response.content))
    # extracting and saving the contents
    zf.extractall(crspFolder)
    # loading the csv file that was extracted. the first three rows are useless text so we will skip them
    # :TODO:F Sometimes the file is saved as .csv and other times as .CSV.
    FF3factors = pd.read_csv(crspFolder + 'F-F_Research_Data_Factors.CSV', skiprows=3)
    FF3factors.columns = ['dates', 'MKT', 'SMB', 'HML', 'RF']
    # The dataframe has the annual factors appended to the bottom of the dataframe, separated by a row of text. We need to
    # find where this occurs and then slice the dataframe to keep everything before it.
    # Uses the isnull function to create a boolean mask that is True where the 'MKT' column is NaN and False
    # otherwise. The idxmax function is then used to return the index label of the first occurrence of the maximum value,
    # i.e. the firt time a 1 occors, representing the first time NaN occurs. Note, I'm using MKT instead of dates because
    # the dates column has useless text followed by NaN in the next column.
    e = FF3factors['MKT'].isnull().idxmax()
    # Next we want to slice the dataframe
    FF3factors = FF3factors.iloc[:e, :]
    # Save the FF dates TODO:Q Matlab code saves ffdates as just a copy of the preexisting dates file. But ffdates start at 192607 instead of 192512.
    FF3factors.dates.to_csv(crspFolder + 'ffdates.csv')
    # Intersect our dates with the ones from Ken French Website; ia is index of elements in dates array in common with those
    # in FF.dates and vica versa.
    _, ia, ib = np.intersect1d(np.array(dates).astype(int), FF3factors['dates'].values.astype(int), return_indices=True)

    #create the variables
    mkt = np.full(len(dates), np.nan)
    smb = np.full(len(dates), np.nan)
    hml = np.full(len(dates), np.nan)
    rf = np.full(len(dates), np.nan)

    # first need to change the datatypes from string to float
    FF3factors = FF3factors.astype(float)
    mkt[ia] = FF3factors.loc[ib, 'MKT'] / 100
    smb[ia] = FF3factors.loc[ib, 'SMB'] / 100
    hml[ia] = FF3factors.loc[ib, 'HML'] / 100
    rf[ia] = FF3factors.loc[ib, 'RF'] / 100

    "Now repeat the steps for UMD factor"
    url = 'http://mba.tuck.dartmouth.edu/pages/faculty/ken.french/ftp/F-F_Momentum_Factor_CSV.zip'

    # tricks the host site into thinking the HTTP call is from user instead of machine
    headers = {"User-Agent": "Mozilla/5.0"}
    response = requests.get(url, headers=headers)
    # saving the zip file as an object
    zf = zipfile.ZipFile(BytesIO(response.content))
    # extracting and saving the contents
    zf.extractall(crspFolder)
    # loading the csv file that was extracted. the first 13 rows are useless text so we will skip them
    UMDFactor = pd.read_csv(crspFolder + 'F-F_Momentum_Factor.CSV', skiprows=13)
    UMDFactor.columns = ['dates', 'UMD']
    # The dataframe has the annual factors appended to the bottom of the dataframe, separated by a row of text. We need to
    # find where this occurs and then slice the dataframe to keep everything before it.
    # Uses the isnull function to create a boolean mask that is True where the 'MKT' column is NaN and False
    # otherwise. The idxmax function is then used to return the index label of the first occurrence of the maximum value,
    # i.e. the firt time a 1 occurs, representing the first time NaN occurs. Note, I'm using MKT instead of dates because
    # the dates column has useless text followed by NaN in the next column.
    e = UMDFactor['UMD'].isnull().idxmax()
    # Next we want to slice the dataframe
    UMDFactor = UMDFactor.iloc[:e, :]

    # Intersect our dates with the ones from Ken French Website; ia is index of elements in dates array in common with those
    # in FF.dates and vica versa.
    _, ia, ib = np.intersect1d(np.array(dates).astype(int), UMDFactor['dates'].values.astype(int), return_indices=True)

    #create the variables
    umd = np.full(len(dates), np.nan)

    # first need to change the datatypes from string to float
    UMDFactor = UMDFactor.astype(float)
    umd[ia] = UMDFactor.loc[ib, 'UMD'] / 100

    "Now repeat the steps for 5 factors"
    url = 'http://mba.tuck.dartmouth.edu/pages/faculty/ken.french/ftp/F-F_Research_Data_5_Factors_2x3_CSV.zip'

    # tricks the host site into thinking the HTTP call is from user instead of machine
    headers = {"User-Agent": "Mozilla/5.0"}
    response = requests.get(url, headers=headers)
    # saving the zip file as an object
    zf = zipfile.ZipFile(BytesIO(response.content))
    # extracting and saving the contents
    zf.extractall(crspFolder)
    # loading the csv file that was extracted. the first 3 rows are useless text so we will skip them
    FF5factors = pd.read_csv(crspFolder + 'F-F_Research_Data_5_Factors_2x3.csv', skiprows=3)
    FF5factors.columns = ['dates', 'MKT', 'SMB', 'HML', 'RMW', 'CMA', 'RF']
    # The dataframe has the annual factors appended to the bottom of the dataframe, seperated by a row of text. We need to
    # find where this occurs and then slice the dataframe to keep everything before it.
    # Uses the isnull function to create a boolean mask that is True where the 'MKT' column is NaN and False
    # otherwise. The idxmax function is then used to return the index label of the first occurrence of the maximum value,
    # i.e. the firt time a 1 occors, representing the first time NaN occurs. Note, I'm using MKT instead of dates because
    # the dates column has useless text followed by NaN in the next column.
    e = FF5factors['MKT'].isnull().idxmax()
    # Next we want to slice the dataframe
    FF5factors = FF5factors.iloc[:e, :]

    # Intersect our dates with the ones from Ken French Website; ia is index of elements in dates array in common with those
    # in FF.dates and vica versa.
    _, ia, ib = np.intersect1d(np.array(dates).astype(int), FF5factors['dates'].values.astype(int), return_indices=True)

    #create the variables
    smb2 = np.full(len(dates), np.nan)
    rmw = np.full(len(dates), np.nan)
    cma = np.full(len(dates), np.nan)

    # first need to change the datatypes from string to float
    FF5factors = FF5factors.astype(float)
    smb2[ia] = FF5factors.loc[ib, 'SMB'] / 100
    rmw[ia] = FF5factors.loc[ib, 'RMW'] / 100
    cma[ia] = FF5factors.loc[ib, 'CMA'] / 100

    # Create and save useful dataframes
    const = np.full(len(rf), 0.01)
    const = pd.Series(const)
    mkt = pd.Series(mkt)
    smb = pd.Series(smb)
    hml = pd.Series(hml)
    rf = pd.Series(rf)
    umd = pd.Series(umd)
    smb2 = pd.Series(smb2)
    rmw = pd.Series(rmw)
    cma = pd.Series(cma)

    factors = ['const', 'mkt', 'smb', 'hml', 'rf', 'umd', 'smb2', 'rmw', 'cma']

    # :TODO Save ['ffdates', 'const', 'rf', 'mkt', 'smb', 'smb2', 'hml', 'umd', 'rmw', 'cma', 'ff3', 'ff4', 'ff5', 'ff6'] in a giant matrix called ff.csv
    ff = pd.concat([dates, const, mkt, smb, hml, rf, umd, smb2, rmw, cma], axis=1)
    ff.columns = ['dates', 'const', 'mkt', 'smb', 'hml', 'rf', 'umd', 'smb2', 'rmw', 'cma']
    ff.to_csv(crspFolder + 'ff.csv')

    ff3 = pd.concat([dates, const, mkt, smb, hml], axis=1)
    ff3.columns = ['dates', 'const', 'mkt', 'smb', 'hml']
    ff3.to_csv(crspFolder + 'ff3.csv')

    ff4 = pd.concat([ff3, umd], axis=1)
    ff4.columns = ['dates', 'const', 'mkt', 'smb', 'hml', 'umd']
    ff4.to_csv(crspFolder + 'ff4.csv')

    ff5 = pd.concat([dates, const, mkt, smb2, hml, rmw, cma], axis=1)
    ff5.columns = ['dates', 'const', 'mkt', 'smb2', 'hml', 'rmw', 'cma']
    ff5.to_csv(crspFolder + 'ff5.csv')

    ff6 = pd.concat([ff5, umd], axis=1)
    ff6.columns = ['dates', 'const', 'mkt', 'smb2', 'hml', 'rmw', 'cma', 'umd']
    ff6.to_csv(crspFolder + 'ff6.csv')

    return




import numpy as np
from scipy.stats import truncnorm, multivariate_normal
from scipy.linalg import sqrtm


def q_draw(p, q, c, varu):
    # Check for length mismatch and scalar requirement
    if len(p) != len(q):
        raise ValueError("p and q are of different lengths.")
    if not np.isscalar(c) or not np.isscalar(varu):
        raise ValueError("c and varu should be scalars.")

    # Initialize
    n_obs = len(p)
    q_nonzero = q != 0
    n_skip = 2
    modp = np.mod(range(1, n_obs+1), n_skip)
    ru = np.random.uniform(size=n_obs)
    # print(ru)
    # Create doubled variables
    p2 = np.column_stack((p, p))
    q = np.column_stack((q, q))

    # Iterate over skip values
    for i_start in range(n_skip):
        # Identify indices to update in this iteration
        k = modp == i_start
        jnz = np.where(k & q_nonzero)[0]

        if len(jnz) > 0:
            # Temporary q for calculations
            # temp_q = np.copy(q2)
            q[jnz, 0] = 1
            q[jnz, 1] = -1

            # Compute u
            cq = c * q
            v = p2 - cq
            u = v[1:, :] - v[:-1, :]
            # print(f'u = {u}')

            # Compute s and log odds
            s = (u ** 2) / (2 * varu)
            # print(f"s = {s}")
            s_sum = s[:-1, :] + s[1:, ]
            # The next 2 lines have the effect of adding a row of 0s to the top and bottom of s before summing.
            s_sum = np.insert(s_sum, 0, s[0], axis=0)
            s_sum = np.append(s_sum, [s[-1]], axis=0)
            s_sum_reduced = s_sum[jnz]

            log_odds = s_sum_reduced[:, 1] - s_sum_reduced[:, 0]

            # Prevent overflow in exp
            log_odds[log_odds > 500] = 500
            odds = np.exp(log_odds)
            p_buy = odds / (1 + odds)

            # Draw new q based on probabilities
            qknz = 1 - 2 * (ru[jnz] > p_buy)
            q[jnz, 0] = qknz
            if i_start < n_skip-1:
                q[jnz, 1] = qknz
    q = q[:, 0]
    return q

def q_draw_vec(p, q, c, varu):
    # Validate the lengths of the input arrays
    if len(p) != len(q):
        raise ValueError("p and q are of different lengths.")

    # Check that c and varu are either scalars or arrays of appropriate lengths
    if not (np.isscalar(c) or len(c) == len(p)) or not (np.isscalar(varu) or len(varu) == len(p) - 1):
        raise ValueError("Invalid dimensions for c or varu.")

    # Create a boolean array indicating where q is not zero
    q_nonzero = q != 0

    # Duplicate the price array p and the trade direction array q
    p2 = np.tile(p, (2, 1)).T
    q_double = np.tile(q, (2, 1)).T

    # Define a constant for the number of elements to skip in the loop
    n_skip = 2

    # Create an array for modulo calculations to determine the update pattern
    mod_p = np.mod(np.arange(len(p)), n_skip)

    for i_start in range(n_skip):

        # Calculate indices for updating q based on the current iteration and nonzero elements in q
        k = mod_p == i_start
        jnz = np.where(k & q_nonzero)[0]

        if len(jnz) > 0:
            # Update the elements of q_double to be alternately 1 and -1 for the indices in jnz
            q_double[jnz, 0] = 1
            q_double[jnz, 1] = -1

            # Calculate cq based on whether c is a scalar or a vector
            if np.isscalar(c):
                cq = c * q_double
            else:
                c2 = np.vstack((c, c)).T
                cq = c2 * q_double

            # Compute the difference between doubled prices and cq
            v = p2 - cq
            u = v[1:] - v[:-1]

            # Calculate s based on whether varu is a scalar or a vector
            if np.isscalar(varu):
                s = (u ** 2) / (2 * varu)
            else:
                varu2 = 2 * np.vstack((varu, varu)).T
                s = (u ** 2) / varu2

            # Extend s to have the same length as p
            s_extended = np.vstack([s, np.zeros((len(p2) - len(s), 2))])

            # Calculate s_sum_all by combining s_extended with itself
            s_sum_all = np.vstack((s_extended[:, 0], np.zeros(len(p2)))).T + np.vstack(
                (np.zeros(len(p2)), s_extended[:, 1])).T

            # Select the relevant rows from s_sum_all
            s_sum = s_sum_all[jnz, :]
            log_odds = s_sum[:, 1] - s_sum[:, 0]

            # Adjust log odds to avoid overflow and calculate the odds
            log_okay = log_odds < 500
            odds = np.exp(log_odds * log_okay)

            # Calculate the probability of buying and adjust it based on log_okay
            p_buy = odds / (1 + odds)
            p_buy = p_buy * log_okay

            # Generate random uniform values and calculate qknz
            ru = np.random.uniform(size=len(p_buy))
            qknz = 1 - 2 * (ru > p_buy)

            # Update q at the indices specified in jnz
            q[jnz] = qknz

    # Return the updated q array
    return q


def bayes_regression_update(prior_mu, prior_cov, y, X, d_var):
    if prior_mu.shape[1] != 1:
        raise ValueError("priorMu should be a column vector")
    if X.shape[0] < X.shape[1]:
        raise ValueError(f"X has fewer rows than columns with dimensions {X.shape[0]}x{X.shape[1]}")
    if X.shape[0] != y.shape[0] or y.shape[1] != 1:
        raise ValueError("Dimensions of X and y are not compatible")
    if prior_mu.shape[0] != X.shape[1]:
        raise ValueError("priorMu and X are not conformable")
    if prior_cov.shape[0] != prior_cov.shape[1] or prior_cov.shape[0] != prior_mu.shape[0]:
        raise ValueError("Dimensions of priorCov are not compatible with priorMu")

    # Create a boolean mask for NaN values
    nan_mask_y = np.isnan(y).flatten()  # Flatten y to make it a 1D array for compatibility
    nan_mask_X = np.isnan(X).any(axis=1)  # Check for NaN in any column of X for each row

    # Combine masks to get rows with NaN in either y or X
    combined_mask = nan_mask_y | nan_mask_X

    # Step 2: Filter out rows with NaN from both y and X
    y = y[~combined_mask]
    X = X[~combined_mask]

    # print(f"d_var = {d_var}")
    # print(f"X = {X}")
    # print(f"y = {y}")
    cov_inv = np.linalg.inv(prior_cov)
    # print(f"cov_inv = {cov_inv}")
    Di = (1 / d_var) * np.dot(X.T, X) + cov_inv
    # print(f"Di = {Di}")
    D = np.linalg.inv(Di)
    dd = (1 / d_var) * np.dot(X.T, y) + np.dot(cov_inv, prior_mu)
    # print(f"dd = {dd}")
    post_mu = np.dot(D, dd)
    post_cov = D

    return post_mu, post_cov


def bayes_variance_update(prior_alpha, prior_beta, u):
    """
    Updates the posterior alpha and beta parameters for an inverted gamma distribution.

    Parameters:
    prior_alpha (float): The prior alpha parameter of the inverted gamma distribution.
    prior_beta (float): The prior beta parameter of the inverted gamma distribution.
    u (numpy.ndarray): Vector of estimated disturbances.

    Returns:
    post_alpha (float): Updated posterior alpha parameter.
    post_beta (float): Updated posterior beta parameter.
    """
    u_cleaned = u[~np.isnan(u)]

    # Update the posterior alpha parameter
    post_alpha = prior_alpha + len(u_cleaned) / 2

    # Update the posterior beta parameter
    post_beta = prior_beta + np.sum(u_cleaned ** 2) / 2

    return post_alpha, post_beta


def rand_std_norm_t(zlow, zhigh):
    PROBNLIMIT = 6
    eps = 100 * np.finfo(float).eps # A small epsilon value

    if zlow == float('-inf') and zhigh == float('inf'):
        return np.random.normal()
    if zlow > PROBNLIMIT and (zhigh == float('inf') or zhigh > PROBNLIMIT):
        return zlow + 100 * eps
    if zhigh < -PROBNLIMIT and (zlow == float('-inf') or zlow < -PROBNLIMIT):
        return zhigh - 100 * eps

    a, b = zlow, zhigh
    # print(f"a, b = {a.dtype}, {b.dtype}")
    if zlow == float('-inf'):
        a = -np.inf
    if zhigh == float('inf'):
        b = np.inf

    if not a<=b:
        raise ValueError(f"Invalid bounds for truncnorm: a={a}, b={b}")
    return truncnorm.rvs(a, b, loc=0, scale=1)


def mvnrnd_t(mu, cov, v_lower, v_upper):
    # f = sqrtm(cov).T  # I'm pretty sure this is incorrect.
    f = np.linalg.cholesky(cov).T
    n = np.prod(mu.shape)
    eta = np.zeros(n)
    # print(f"f = sqrtm(cov) = {f}")
    # print(f"eta = np.zeros(n) = {eta}")


    for k in range(n):
        etasum = np.dot(f[k, :k], eta[:k])
        # print(f"etasum = {etasum}")
        low = (v_lower[k] - mu[k] - etasum) / f[k, k]
        high = (v_upper[k] - mu[k] - etasum) / f[k, k]
        eta[k] = rand_std_norm_t(low, high)

    return mu + np.dot(f, eta)


def roll_gibbs_beta(p, pm, q, n_sweeps, reg_draw, varu_draw, q_draw_bool, varu_start, c_start, beta_start):
    """
    Perform Gibbs sampling to estimate parameters in the Roll model.

    Parameters:
    p (numpy.ndarray): Vector of trade prices.
    pm (numpy.ndarray): Vector of mid prices.
    q (numpy.ndarray): Vector of trade directions.
    n_sweeps (int): Number of Gibbs sampling iterations.
    reg_draw (bool): Flag to perform regression draw.
    varu_draw (bool): Flag to perform variance draw.
    q_draw_bool (bool): Flag to perform q draw.
    varu_start (float): Initial value of variance.
    c_start (float): Initial value of cost parameter c.
    beta_start (float): Initial value of beta.
    print_level (int): Verbosity level for printing diagnostics.

    Returns:
    numpy.ndarray: Output matrix with parameters from each sweep. Columns are c, beta, and varu.
    """
    n_obs = len(p)

    # Check for length mismatch
    if len(q) != n_obs or len(pm) != n_obs:
        print('RollGibbsBeta length mismatch')
        return None

    # Calculate price change
    dp = p[1:] - p[:-1]

    # Initialize q based on price sign changes if required
    if q_draw_bool:
        q_initial = np.sign(dp)
        q_initial = np.insert(q_initial, 0, 1)  # Insert 1 at beginning to match length of q
        q = np.where(q != 0, q_initial, q)

    # Initialize variance, cost, and beta parameters
    # print(f"varu_start = {varu_start}")
    varu = max(varu_start, 0.001)
    c = max(c_start, 0.01)
    beta = max(beta_start, 1)

    # Output matrix initialization
    parm_out = np.zeros((n_sweeps, 3))

    for sweep in range(n_sweeps):
        # print(f"Sweep {sweep + 1}/{n_sweeps}")
        # Calculate changes in trade directions and mid prices
        dq = q[1:] - q[:-1]
        dpm = pm[1:] - pm[:-1]
        # Perform regression draw if enabled
        if reg_draw:
            # varu = max(varu_start, 0.001)
            prior_mu = np.array([[0], [1]])  # Prior mean
            prior_cov = np.diag([1, 2])  # Prior covariance
            X = np.column_stack((dq, dpm))  # Design matrix
            # print(f"X has dimensions {X.shape}")
            # print(f'varu = {varu}')
            post_mu, post_cov = bayes_regression_update(prior_mu, prior_cov, dp.reshape(-1, 1), X, varu)
            # print(f"Post cov = {post_cov}")
            coeff_lower = np.array([0, float('-inf')])
            coeff_upper = np.array([float('inf'), float('inf')])
            coeff_draw = mvnrnd_t(post_mu.flatten(), post_cov, coeff_lower, coeff_upper)
            c, beta = coeff_draw[0], coeff_draw[1]  # Extract c and beta

        # Perform variance draw if enabled
        if varu_draw:
            u = dp - c * dq[:len(dp)] - beta * dpm[:len(dp)]  # Calculate disturbances
            prior_alpha = 1e-12
            prior_beta = 1e-12
            post_alpha, post_beta = bayes_variance_update(prior_alpha, prior_beta, u)
            varu = 1 / np.random.gamma(post_alpha, 1 / post_beta)
            # print(f"Sweep {sweep}, varu={varu}, post_alpha={post_alpha}, post_beta={post_beta}")

        # Update q if required
        if q_draw_bool:
            p2 = p - beta * pm
            q = q_draw(p2, q, c, varu)

        # Store parameters in output array
        parm_out[sweep, :] = [c, beta, varu]

    return parm_out

#
# "Testing bayes_regression_update and roll_gibbs_beta"
# # Example test data
# np.random.seed(0)  # for reproducibility
# n_samples = 100
# n_features = 2
#
# X_test = np.random.randn(n_samples, n_features)  # Design matrix
# y_test = np.random.randn(n_samples, 1) * 0.5  # Dependent variable with some noise
#
# prior_mu_test = np.zeros((n_features, 1))  # Prior mean (zero)
# prior_cov_test = np.eye(n_features)  # Prior covariance (identity matrix)
# d_var_test = 0.5  # Variance of error
#
# # Run bayes_regression_update
# post_mu, post_cov = bayes_regression_update(prior_mu_test, prior_cov_test, y_test, X_test, d_var_test)
#
# # Print or inspect the results
# print("Posterior Mean (post_mu):", post_mu)
# print("Posterior Covariance (post_cov):", post_cov)
#
# # Example test data for roll_gibbs_beta
# p_test = np.random.normal(size=100)
# pm_test = np.random.normal(size=100)
# q_test = np.random.choice([-1, 1], size=100)
#
# # Parameters for the Gibbs sampler
# n_sweeps_test = 1000
# reg_draw_test = True
# varu_draw_test = True
# q_draw_test = True
# varu_start_test = 0.001
# c_start_test = 0.01
# beta_start_test = 1
# print_level_test = 1
#
# # Run roll_gibbs_beta
# results = roll_gibbs_beta(p_test, pm_test, q_test, n_sweeps_test, reg_draw_test, varu_draw_test, q_draw_test, varu_start_test, c_start_test, beta_start_test, print_level_test)
#
# # Inspect results
# print("Results from RollGibbsBeta:", results)
#

"To test q_draw"

# # Example test data
# # Define p array
# p = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
#               11, 12, 13, 14, 15, 16, 17, 18, 19, 20,
#               21, 22, 23, 24, 25, 26, 27, 28, 29, 30,
#               31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
#               41, 42, 43, 44, 45, 46, 47, 48, 49, 50,
#               51, 52, 53, 54, 55, 56, 57, 58, 59, 60,
#               61, 62, 63, 64, 65, 66, 67, 68, 69, 70,
#               71, 72, 73, 74, 75, 76, 77, 78, 79, 80,
#               81, 82, 83, 84, 85, 86, 87, 88, 89, 90,
#               91, 92, 93, 94, 95, 96, 97, 98, 99, 100])
#
# # Define q array
# q = np.array([-1, 1, -1, -1, 1, -1, 1, 1, -1, 1,
#               -1, 1, -1, 1, 1, 1, 1, 1, 1, 1,
#               1, -1, -1, -1, -1, 1, 1, -1, -1, -1,
#               1, 1, 1, -1, 1, 1, 1, 1, 1, -1,
#               1, 1, -1, 1, 1, 1, 1, -1, 1, 1,
#               1, 1, -1, -1, -1, 1, -1, 1, 1, 1,
#               1, -1, -1, -1, -1, -1, -1, -1, 1, -1,
#               -1, -1, 1, -1, 1, -1, 1, -1, -1, 1,
#               1, 1, -1, 1, -1, 1, 1, -1, -1, 1,
#               -1, -1, -1, -1, 1, 1, -1, -1, 1, -1])
#
# ru = np.array([
#     0.7223837, 0.1345663, 0.4886556, 0.1905619, 0.968369, 0.1272038, 0.8160228, 0.5720403, 0.8358008, 0.9459323,
#     0.5218109, 0.3258735, 0.7206986, 0.8455685, 0.3894503, 0.5619935, 0.9888516, 0.4692584, 0.7506123, 0.3037273,
#     0.685163, 0.7390812, 0.734625, 0.7475037, 0.6117248, 0.7370916, 0.1971345, 0.3631171, 0.6971642, 0.9983765,
#     0.0003153, 0.6928824, 0.5568195, 0.6228948, 0.6122706, 0.5200895, 0.1128145, 0.80839, 0.1276258, 0.7178436,
#     0.8036793, 0.4268984, 0.7352105, 0.9730008, 0.320453, 0.2439898, 0.9974192, 0.3417673, 0.3071908, 0.5365895,
#     0.267439, 0.7800667, 0.0860517, 0.9082585, 0.5879633, 0.9047471, 0.6610216, 0.5806289, 0.5081874, 0.8883773,
#     0.4213902, 0.7945829, 0.8044851, 0.9464459, 0.2317895, 0.5462842, 0.5559665, 0.6792165, 0.47886, 0.964277,
#     0.7418761, 0.8125367, 0.2915919, 0.4914265, 0.4479994, 0.4391654, 0.0639266, 0.5267615, 0.2011821, 0.2892307,
#     0.1991538, 0.0163052, 0.7327378, 0.721235, 0.4327409, 0.9739116, 0.1942334, 0.5365691, 0.7504959, 0.7876037,
#     0.5749443, 0.8453409, 0.4943481, 0.4200899, 0.263467, 0.1557632, 0.3722185, 0.7363796, 0.582786, 0.6489519
# ])
#
# c = 0.5  # Scalar value
# varu = 2.0  # Scalar value
# print_level = 0  # Verbosity level
#
# q_updated = q_draw(p, q, c, varu, print_level, ru)



import numpy as np
import pandas as pd
from AssayingAnomalies.Functions.estFactorRegs import estFactorRegs
from AssayingAnomalies.Functions.prt_sort_results import prt_sort_results
from AssayingAnomalies.Functions.runUnivSort import runUnivSort
from AssayingAnomalies.Functions.grs_test_p import grs_test_p


def runBivSort(params, ret, ind, nptf1, nptf2, mcap, dates, **kwargs):
    # Convert input DataFrames to Numpy array
    if isinstance(ret, pd.DataFrame):
        ret = ret.to_numpy()

    if isinstance(ind, pd.DataFrame):
        ind = ind.to_numpy()

    if isinstance(mcap, pd.DataFrame):
        mcap = mcap.to_numpy()

    if isinstance(dates, pd.DataFrame):
        dates = dates.values.flatten()

    # Default parameter values
    p = {
        'factorModel': 4,
        'addLongShort': 1,
        'printResults': 1,
        'plotFigure': 0,
        'timePeriod': [dates[0], dates[-1]],
        'holdingPeriod': 1,
        'weighting': 'v',
        'tcosts': -99,
    }

    # Update parameters with user-specified values (if provided).
    for key, value in kwargs.items():
        if key in p:
            p[key] = value

    # Check if weighting is correctly specified
    expected_weighting = {'V', 'v', 'E', 'e'}
    if p['weighting'] not in expected_weighting:
        raise ValueError("Invalid weighting value. Must be one of {'V', 'v', 'E', 'e'}")

    # Update parameters with user-specified values (if provided).
    for key, value in kwargs.items():
        if key in p:
            p[key] = value

    # Run a univariate sort to get the underlying portfolios
    res = runUnivSort(params=params, ret=ret, ind=ind, mcap=mcap, dates=dates, **p)

    # Add the GRS test results to 'res' dictionary
    res["grsPval"], res["grsFstat"], res["grsDoF"] = grs_test_p(res)

    # Store a few constants
    nMonths = len(res["dates"])

    # Conditional strategies: High minus low var 1, conditioned on var 2
    longShortCondStrats = np.empty((nMonths, nptf2))
    longShortCondStrats[:] = np.nan
    for i in range(nptf2):
        thisLongInd = (nptf1 - 1) * nptf2 + i
        thisShortInd = i
        longShortCondStrats[:, i] = res["pret"][:, thisLongInd] - res["pret"][:, thisShortInd]

    cond_res_1 = estFactorRegs(params, longShortCondStrats, res["dates"], p["factorModel"], addLongShort=0, inputIsExcessRets=1)

    # Conditional strategies: High minus low var 2, conditioned on var 1
    longShortCondStrats = np.empty((nMonths, nptf1))
    longShortCondStrats[:] = np.nan
    for i in range(nptf1):
        thisLongInd = (i * nptf2) + nptf2 - 1
        thisShortInd = (i * nptf2)
        longShortCondStrats[:, i] = res["pret"][:, thisLongInd] - res["pret"][:, thisShortInd]

    cond_res_2 = estFactorRegs(params, longShortCondStrats, res["dates"], p["factorModel"], addLongShort=0, inputIsExcessRets=1)

    # Update cond_res with the second set of conditional results
    cond_res = {'cond_res_1': cond_res_1, 'cond_res_2': cond_res_2}

    if p['printResults'] != 0:
        # Print the GRS results
        grsres1 = f"{res['grsPval'][0]} {res['grsFstat'][0]} {round(res['grsDoF'][0][0])} {round(res['grsDoF'][0][1])}"
        grsres2 = f"{res['grsPval'][1]} {res['grsFstat'][1]} {round(res['grsDoF'][1][0])} {round(res['grsDoF'][1][1])}"
        print("      GRS test results:  p-value     F-stat  df1  df2")
        print(f"      full test -        {grsres1}")
        print(f"      partial test -     {grsres2}")

        # # Add a few descriptors to the structure
        for i in range(2):
            cond_res[f'cond_res_{i+1}']['w'] = p['weighting']
            cond_res[f'cond_res_{i+1}']['hperiod'] = p['holdingPeriod']


        # # Print the conditional strategies
        prt_sort_results(cond_res['cond_res_1'], lsprt=0)
        prt_sort_results(cond_res['cond_res_2'], lsprt=0)

        # Print the average returns, number of firms, and market cap
        print("Portfolio Average Excess Returns (%/month)")
        print("     ")
        if res['xret'].shape[0] == nptf1 * nptf2: # adding this to check if the l/s port xret has been appended or not
            print(np.round(res['xret'], decimals=4).reshape(nptf2, nptf1))
        else:
            print(np.round(res['xret'][:-1], decimals=4).reshape(nptf2, nptf1))
        print("Portfolio Average Number of Firms")
        print("     ")
        if res['xret'].shape[0] == nptf1 * nptf2:
            print(np.round(np.nanmean(res['ptfNumStocks'], axis=0), decimals=0).reshape(nptf2, nptf1))
        else:
            print(np.round(np.nanmean(res['ptfNumStocks'], axis=0), decimals=0)[:-1].reshape(nptf2, nptf1))
        print("Portfolio Average Firm Size ($10^6)")
        print("     ")
        if res['xret'].shape[0] == nptf1 * nptf2:
            print(np.round(np.nanmean(res['ptfMarketCap'] / res['ptfNumStocks'], axis=0), decimals=0).reshape(nptf2, nptf1))
        else:
            print(np.round(np.nanmean(res['ptfMarketCap'] / res['ptfNumStocks'][:, :-1], axis=0), decimals=0).reshape(nptf2, nptf1))

    return res, cond_res

# # Test the function
# import scipy.io
# import os
# from AssayingAnomalies import Config
# from AssayingAnomalies.Functions.makeBivSortInd import makeBivSortInd
#
# params = Config()
# params.prompt_user()
# params.make_folders()
# path = r"C:\Users\josh\OneDrive - University at Buffalo\Desktop\Spring_2023\AssayingAnomalies-main\AssayingAnomalies-main\Data" + os.sep
# R = scipy.io.loadmat(path + 'R.mat')['R']
# # R = pd.read_csv(path + 'R.csv', index_col=0)
# me = scipy.io.loadmat(path + 'me.mat')['me']
# NYSE = scipy.io.loadmat(path + 'NYSE.mat')['NYSE']
# ind1 = makeBivSortInd(me, 5, R, 5)
# ret = scipy.io.loadmat(path + 'ret.mat')['ret']
# dates = scipy.io.loadmat(path + os.sep + 'CRSP' + os.sep + 'dates.mat')['dates']
# dates = dates.flatten()
# test_res1, test_cond_res1 = runBivSort(params=params, ret=ret, ind=ind1, nptf1=5, nptf2=5, dates=dates, mcap=me)
#
# ind2 = makeBivSortInd(me, 2, R, [30, 70])
# test_res2, test_cond_res2 = runBivSort(params=params, ret=ret, ind=ind2, nptf1=3, nptf2=2, dates=dates, mcap=me)
#
# ind3 = makeBivSortInd(me, 5, R, 5, sort_type='conditional')
# test_res3, test_cond_res3 = runBivSort(params=params, ret=ret, ind=ind3, nptf1=5, nptf2=5, dates=dates, mcap=me)



from .assignToPtf import assignToPtf
from .calcPtfRets import calcPtfRets
from .calcTcosts import calcTcosts
from .crspGibbsBuildv01 import *
# from .create_pgpass_file import create_pgpass_file
from .estFactorRegs import estFactorRegs
from .fill_months import fill_months
from .fill_missing_tcosts import fill_missing_tcosts
from .getCompustatData import getCOMPUSTATData
from .getCRSPData import getCRSPData
from .get_crsp_daily_data import get_crsp_daily_data
from .getFFDailyFactors import getFFDailyFactors
from .getFFFactors import getFFFactors
from .getFredData import getFredData
from .Gibbs import *
from .grs_test_p import grs_test_p
from .makeAbdiRanaldi import makeAbdiRanaldi
from .makeBetas import makeBetas
from .makeCOMPUSTATDerivedVariables import makeCOMPUSTATDerivedVariables
from .makeCOMPUSTATVariables import makeCOMPUSTATVariables
from .makeCorwinSchultz import makeCorwinSchultz
from .makeCRSPDailyData import makeCRSPDailyData
from .makeCRSPDailyDerivedVariables import makeCRSPDailyDerivedVariables
from .makeCRSPDerivedVariables import makeCRSPDerivedVariables
from .makeCRSPMonthlyData import makeCRSPMonthlyData
from .makeFF10Indus import makeFF10Indus
from .makeFF17Indus import makeFF17Indus
from .makeFF49Indus import makeFF49Indus
from .makeGibbs import makeGibbs
from .make_hf_effective_spreads import get_hf_spreads_data
from .make_hf_effective_spreads import make_hf_effective_spreads
from .make_hf_effective_spreads import extend_hf_effective_spreads
from .makeIndustryReturns import makeIndustryReturns
from .makeIndustryClassifications import makeIndustryClassifications
from .makeKyleObizhaeva import makeKyleObizhaeva
from .makeNovyMarxVelikovAnomalies import makeNovyMarxVelikovAnomalies
from .makePastPerformance import makePastPerformance
from .makeUniverses import makeUniverses
from .makeUnivSortInd import makeUnivSortInd
from .makeBivSortInd import makeBivSortInd
from .makeTradingCosts import makeTradingCosts
from .mergeCRSPCOMP import mergeCRSPCOMP
from .prt_FMB_results import prt_FMB_results
from .prt_sort_results import prt_sort_results
from .rank_with_nan import rank_with_nan
from .remote_get_compustat_data import remote_get_compustat_data
from .remote_get_crsp_data import remote_get_crsp_data
from .remote_get_daily_crsp import remote_get_daily_crsp
from .remote_get_dsi import remote_get_dsi
from .run_Fama_MacBeth import run_Fama_MacBeth
from .runUnivSort import runUnivSort
from .runBivSort import runBivSort



import pandas as pd
import os
import numpy as np
from datetime import datetime
from concurrent.futures import ThreadPoolExecutor
import dask.dataframe as dd


def makeCRSPDailyData(params):

    # Timekeeping
    print(f"\nNow working on making CRSP daily data. Run started at {datetime.now()}.\n")

    # Store the daily CRSP data path
    daily_crsp_path = params.daily_crsp_folder + os.sep

    # If files are .csv then change to .parquet
    # for file in daily_crsp_files:
    #     temp_df = pd.read_csv(file)
    #     parquet_file = file.replace('.csv', '.parquet')
    #     temp_df.to_parquet(parquet_file)

    # Load the permnos and create ddates
    permno = pd.read_csv(params.crspFolder + os.sep + 'permno.csv', index_col=0).astype(str)  # Assuming permno is in a CSV file
    ddates = dd.read_parquet(daily_crsp_path + 'crsp_dsf*.parquet', columns='date').unique()
    ddates = ddates.compute()

    # Load the var_names for individual dataframes. The first 2 col names are unnamed, date, and permno, so we
    # exclude them.
    var_names = pd.read_parquet(daily_crsp_path + 'crsp_dsf_' + str(params.sample_end) + '.parquet').columns.values[2:].flatten()

    # Instead of using pd.read_csv, use dd.read_parquet to load the Parquet files
    # Dask will handle these files as a single DataFrame, but under the hood,
    # it will process them in parallel, in small chunks
    if not params.remote_or_not:
        for i, var in enumerate(var_names):
            # for i, var in enumerate(['ret']):
            crsp_dsf = dd.read_parquet(daily_crsp_path + 'crsp_dsf*.parquet', columns=['date', 'permno', var])
            # crsp_dsf['permno'] = crsp_dsf['permno'].astype(int)
            crsp_dsf = crsp_dsf[crsp_dsf['permno'].isin(permno.values.flatten())]
            # When working with parquet files, it uses None type for missing objects values and nan for missing numeric
            # values. When we try and pivot, the None values are causing errors for some of the permnos.
            replace_dict = {None: np.nan}
            crsp_dsf[var] = crsp_dsf[var].replace(replace_dict)
            crsp_dsf = crsp_dsf.compute()
            temp_df = crsp_dsf.pivot_table(index='date', columns='permno', values=var)
            temp_df.columns = temp_df.columns.values.astype(int).astype(str)
            temp_df = temp_df.reindex(index=ddates, columns=permno.values.flatten(), fill_value=np.nan)
            temp_df.columns = temp_df.columns.astype(str)
            # testing = temp_df.reindex(index=ddates, columns=permno.values.flatten().astype(str), fill_value=np.nan)
            # test = np.intersect1d(ddates, temp_df.index)
            # test2 = np.intersect1d(permno.values.flatten().astype(str), temp_df.columns)
            # print(ddates.dtype, temp_df.index.dtype)
            # print(permno.values.flatten().astype(str).dtype, temp_df.columns.dtype)

            if var in ['ret', 'vol']:
                temp_df.to_parquet(daily_crsp_path + f'd{var}_x_adj.parquet')
                print(f"Finished processing {var}: Variable {i+1}/{len(var_names)} with shape {temp_df.shape}")
            else:
                temp_df.to_parquet(daily_crsp_path + f'd{var}.parquet')
                print(f"Finished processing {var}: Variable {i+1}/{len(var_names)} with shape {temp_df.shape}")

    else:
        def process_and_save_var(permno, ddates, daily_crsp_path, var):
            """
            Process a single variable and save the result to a CSV file.
            """
            try:
                print(f"Processing {var}")
                crsp_dsf = dd.read_parquet(daily_crsp_path + 'crsp_dsf*.parquet', columns=['date', 'permno', var])
                # crsp_dsf['permno'] = crsp_dsf['permno'].astype(int)
                crsp_dsf = crsp_dsf[crsp_dsf['permno'].isin(permno.values.flatten())]
                # When working with parquet files, it uses None type for missing objects values and nan for missing numeric
                # values. When we try and pivot, the None values are causing errors for some of the permnos.
                replace_dict = {None: np.nan}
                crsp_dsf[var] = crsp_dsf[var].replace(replace_dict)
                crsp_dsf = crsp_dsf.compute()
                temp_df = crsp_dsf.pivot_table(index='date', columns='permno', values=var)
                temp_df.columns = temp_df.columns.values.astype(int).astype(str)
                temp_df = temp_df.reindex(index=ddates, columns=permno.values.flatten(), fill_value=np.nan)
                temp_df.columns = temp_df.columns.astype(str)
                # print(ddates.dtype, temp_df.index.dtype)
                # print(permno.values.flatten().astype(str).dtype, temp_df.columns.dtype)
                if var in ['ret', 'vol']:
                    temp_df.to_parquet(daily_crsp_path + f'd{var}_x_adj.parquet')
                else:
                    temp_df.to_parquet(daily_crsp_path + f'd{var}.parquet')
                # temptable.to_csv(daily_crsp_path + f'd{var}.csv')
                print(f"Finished processing {var} with shape {temp_df.shape}")
            except Exception as e:
                print(f"Error processing {var}: {e}")

        with ThreadPoolExecutor(max_workers=params.num_cpus) as executor:
            # Submit tasks to the thread pool
            futures = [executor.submit(process_and_save_var, permno, ddates, daily_crsp_path, var) for var in var_names]
            # futures = [executor.submit(process_and_save_var, permno, ddates, daily_crsp_path, var) for var in ['numtrd', 'openprc']]

            # wait for each task to complete and print any errors if they exist.
            for future in futures:
                future.result()  # Waits for the individual task to complete


    # Create ddates in YYYYMMDD format
    ddates = dd.read_parquet(daily_crsp_path + 'crsp_dsf*.parquet', columns='date').unique()
    ddates = dd.to_datetime(ddates)
    ddates = ddates.dt.strftime('%Y%m%d').astype(int).values.flatten()
    ddates = ddates.compute()

    # Create the end of month flag and store ddates
    yyyy_mm = np.floor(ddates / 100).astype(int)
    eomflag = yyyy_mm != np.roll(yyyy_mm, -1)
    eomflag[-1] = True

    ddates_csv = pd.DataFrame(ddates)
    ddates_csv.to_csv(daily_crsp_path + 'ddates.csv')

    eomflag_csv = pd.DataFrame(eomflag)
    eomflag_csv.to_csv(daily_crsp_path + 'eomflag.csv')

    print(f"Daily data processing completed at {datetime.now()}.\n")

    return




# from AssayingAnomalies.config import Config
# params = Config()
# params.set_up()
# makeCRSPDailyData(params)




import numpy as np


def assignToPtf(x, bPtsMat):
    """
    Function that assigns each firm-month to the bin it belongs to in the
    particular month. It assigns a zero if a firm-month is not held in any
    portfolio.

    Parameters
    ----------
    x : numpy array or pandas DataFrame
        Matrix based on which we want to assign the bins.
    bPtsMat : numpy array
        A matrix with the breakpoints (values) for each time period.
            e.g. np.nanpercentile(me, np.linspace(0, 100, 5), axis=1).T

    Returns
    -------
    ind : numpy array
        A matrix that indicates the bin each stock-time is in.
    """

    # Store several dimensions
    nStocks = x.shape[1]
    nPeriods = x.shape[0]
    nBPoints = bPtsMat.shape[1]

    # Create a matrix with the same dimensions as x
    ind = np.zeros((nPeriods, nStocks))

    # Assign the indicator for the first portfolio
    bPtsPtfOne = bPtsMat[:, 0]
    rptdBPtsPtfOne = np.repeat(bPtsPtfOne[:, np.newaxis], nStocks, axis=1)
    ind[x < rptdBPtsPtfOne] = 1

    # Repeat the same for all the other breakpoints
    for j in range(1, nBPoints + 1):
        bPtsPtfJ = bPtsMat[:, j - 1]
        rptdBPtsPtfJ = np.repeat(bPtsPtfJ[:, np.newaxis], nStocks, axis=1)
        ind[x >= rptdBPtsPtfJ] = j + 1

    return ind


# saveFolder = r'/home/jlaws13/PycharmProjects/AssayingAnomalies_root/Data/CRSP/'
# me = pd.read_csv(saveFolder + 'me.csv', index_col=0).astype(float)
# test_bpts = np.nanpercentile(me, np.linspace(0, 100, 5), axis=1).T
# test_ind = assignToPtf(me, test_bpts)


"""makeCOMPUSTATVariables(parameters, dataframe, quarterlyIndicator) uses the stored annual and quarterly COMPUSTAT-CRSP
linked files to create matrices of dimensions number of months by number of stocks for all variables."""

import numpy as np
import pandas as pd
import os
from datetime import datetime
from concurrent.futures import ThreadPoolExecutor


def makeCOMPUSTATVariables(params, data, quarterly_indicator=False):
    # set paths
    crspFolder = params.crspFolder + os.sep
    compFolder = params.compFolder + os.sep

    # load a few variables
    permno = pd.read_csv(crspFolder + 'permno.csv', dtype=np.int32, index_col=0)
    dates = pd.read_csv(crspFolder + 'dates.csv', dtype=np.int32, index_col=0)
    ret = pd.read_csv(crspFolder + 'ret.csv', index_col=0)

    # ret columns names need to be integers for comparison later on
    data['permno'] = data['permno'].astype(int)
    ret.columns = ret.columns.values.astype(float).astype(int)

    # Store a few constants
    nStocks = len(permno)
    nMonths = len(dates)
    nObs = nStocks * nMonths

    # Create the linking table with CRSP
    # rptdDates = np.tile(dates, (nStocks, 1)).T.flatten()
    # rptdPermno = np.tile(permno, (nMonths, 1)).flatten()
    # crspMatLink = np.vstack((rptdPermno, rptdDates)).T
    # crspMatLinkTab = pd.DataFrame(crspMatLink, columns=['permno', 'd
    # Store the variable names & drop the permno and dates
    varNames = data.columns[1:].values
    idxToDrop = np.isin(varNames, ['permno', 'dates'])
    varNames = varNames[~idxToDrop]

    # Store the number of variable names
    nVarNames = len(varNames)

    # Create dataframes of single varNames
    # test = varNames[2:4]
    # for i, col in enumerate(test):
    # print(data.columns)
    if not params.remote_or_not:
        for i, col in enumerate(varNames):
            print(f"Now working on COMPUSTAT variable {col}, which is {i + 1}/{nVarNames}")
            # temptable = pd.pivot_table(crsp_msf, index='permno', columns='dates', values = i)
            temptable = pd.pivot(data, index='dates', columns='permno', values=col)
            # temptable.to_csv(compFolder + col + '.csv')

            # Align columns and rows with ret
            # Get firms that are in ret but not in compustat variable
            missing_permnos = ret.columns.difference(temptable.columns)

            # Create corresponding columns
            for permno in missing_permnos:
                temptable[permno] = np.nan

            # Align index (dates)
            missing_dates = ret.index.difference(temptable.index)
            for date in missing_dates:
                temptable.loc[date] = np.nan

            # Sort index and columns to match 'ret'
            temptable = temptable.reindex(index=ret.index, columns=ret.columns)
            # print(temptable.shape)

            if quarterly_indicator is True and col != 'FQTR':
                stocks_with_data_ind = temptable.columns[temptable.notna().any().values]
                n_stocks_with_data = len(stocks_with_data_ind)

                # Loop through them
                for c in stocks_with_data_ind:
                    # Find the first and last rows
                    first_r = temptable.loc[:, c].first_valid_index().astype(int)
                    last_r = temptable.loc[:, c].last_valid_index().astype(int)

                    # Loop through the rows/months
                    for r in range(first_r + 1, min(nMonths, last_r + 3)):
                        if np.isnan(temptable.at[r, c]):
                            # Fill in the missing ones
                            temptable.at[r, c] = temptable.at[r - 1, c]

            # Save the  temptable
            temptable.to_csv(compFolder + col.upper() + '.csv')

    else:  # :TODO When aligning columns, use .reindex instead. It is much faster.
        def process_and_save_var(permno, dates, col, quarterly_indicator=False):
            print(f"Now working on COMPUSTAT variable {col}")
            # temptable = pd.pivot_table(crsp_msf, index='permno', columns='dates', values = i)
            temptable = pd.pivot(data, index='dates', columns='permno', values=col)
            # temptable.to_csv(compFolder + col + '.csv')

            # Align columns and rows.
            temptable = temptable.reindex(index=dates.values.flatten(), columns=permno.values.flatten(), fill_value=np.nan)

            # # Align columns and rows with ret
            # # Get firms that are in ret but not in compustat variable
            # missing_permnos = ret.columns.difference(temptable.columns)
            #
            # # Create corresponding columns
            # for permno in missing_permnos:
            #     temptable[permno] = np.nan
            #
            # # Align index (dates)
            # missing_dates = ret.index.difference(temptable.index)
            # for date in missing_dates:
            #     temptable.loc[date] = np.nan
            #
            # # Sort index and columns to match 'ret'
            # temptable = temptable.reindex(index=ret.index, columns=ret.columns)
            # print(temptable.shape)

            if quarterly_indicator is True and col != 'FQTR':
                stocks_with_data_ind = temptable.columns[temptable.notna().any().values]
                n_stocks_with_data = len(stocks_with_data_ind)

                # Loop through them
                for c in stocks_with_data_ind:
                    # Find the first and last rows
                    first_r = temptable.loc[:, c].first_valid_index().astype(int)
                    last_r = temptable.loc[:, c].last_valid_index().astype(int)

                    # Loop through the rows/months
                    for r in range(first_r + 1, min(nMonths, last_r + 3)):
                        if np.isnan(temptable.at[r, c]):
                            # Fill in the missing ones
                            temptable.at[r, c] = temptable.at[r - 1, c]

            # Save the  temptable
            temptable.to_csv(compFolder + col.upper() + '.csv')

            # Print statement
            print(f"Completed COMPUSTAT variable {col} with shape {temptable.shape}")

        with ThreadPoolExecutor(max_workers=params.num_cpus) as executor:
            # Submit tasks to the thread pool
            if quarterly_indicator:
                futures = [executor.submit(process_and_save_var, permno, dates, var, quarterly_indicator=True) for var in varNames]
            else:
                futures = [executor.submit(process_and_save_var, permno, dates, var, quarterly_indicator=False) for var in varNames]
            # wait for each task to complete and print any errors if they exist.
            for future in futures:
                future.result()  # Waits for the individual task to complete

# Uncomment to debug
# from AssayingAnomalies import Config
# params = Config()
# params.set_up()
# data = pd.read_csv(params.crspFolder + os.sep + 'adj_comp_fundq_linked.csv')
# makeCOMPUSTATVariables(params=params, data=data, quarterly_indicator=True)
#
